"""
Utility methods to parse and summarize posts
"""
from datetime import datetime
from statistics import fmean
from smstats.custom import PostDetails

def get_post_details(post):
    """
    Summarize a subset of user posts
    """
    summary = PostDetails()

    # Param 1: Post Length
    summary.length = len(post['message'])

    # Param 2: Post month
    created_time = datetime.fromisoformat(post['created_time'])
    summary.month = f'{created_time.month}-{created_time.year}'

    # Param 3: Post Week
    summary.week = created_time.date().isocalendar().week

    # Param 4: User
    summary.user = post['from_id']

    return summary

def build_posts_summary(summary, posts):
    """
    Incremently build posts summary
    """

    for post in posts:
        details = get_post_details(post)

        # isummary is interval summary
        # Update monthly and weekly stats
        for isummary in [summary.monthly[details.month],
                         summary.weekly[details.week]]:
            # Total posts
            isummary.users[details.user].total += 1

            # Longest post so far
            longest = isummary.users[details.user].longest
            isummary.users[details.user].longest = max(longest, details.length)

            # Total Length
            isummary.users[details.user].sum_length += details.length

def get_stats(summary):
    """
    Generate reports from collected data
    """
    # Will store result in a dictionary. Will be easily translated to json
    stats = {
        'avg_postlen_per_month': {},
        'longest_post_per_month': {},
        'total_posts_by_week': {},
        'avg_post_per_user_per_month': {}
    }

    # isummary is interval summary
    # psummary is post summary
    for month, isummary in summary.monthly.items():
        # Stat 1: Average Length of Posts per month
        total_length = sum([psummary.sum_length for psummary in isummary.users.values()])
        month_posts = sum([psummary.total for psummary in isummary.users.values()])
        stats['avg_postlen_per_month'][month] = round(total_length/month_posts, 2)

        # Stat 2: Longest post per month
        stats['longest_post_per_month'][month] = \
        max([psummary.longest for psummary in isummary.users.values()])

        stats['avg_post_per_user_per_month'][month] = \
        fmean([psummary.total for psummary in isummary.users.values()])

    # isummary is interval summary
    # psummary is post summary
    for week, isummary in summary.weekly.items():
        # Stat 3: Total posts per week
        stats['total_posts_by_week'][str(week)] = \
        sum([psummary.total for psummary in isummary.users.values()])

    return stats



"""
Supermetrics REST service details
"""
from dataclasses import dataclass
from smstats.custom import ExpectedResponse

# Default configuration to register token
TOKEN_URL = 'https://api.supermetrics.com/assignment/register'
TOKEN_REG_REQUEST = {
    'client_id': 'ju16a6m81mhid5ue1z3v2g0uh',
    'email': 'manish@gmail.com',
    'name': 'manish'
}
TOKEN_KEY = ('data', 'sl_token')

# Default configuration to get posts
POSTS_URL = 'https://api.supermetrics.com/assignment/posts'
MAX_POST_PAGE = 10
INVALID_TOKEN_RESPONSE = ExpectedResponse(status=500,
                                          key=('error', 'message'),
                                          value='Invalid SL Token')
POSTS_KEY = ('data', 'posts')

# User editable configuration
@dataclass
class Config:
    """user editable configuration"""
    client_id: str = TOKEN_REG_REQUEST['client_id']
    email: str = TOKEN_REG_REQUEST['email']
    name: str = TOKEN_REG_REQUEST['name']
    max_page: int = MAX_POST_PAGE



"""
Utility function to handle REST response
"""
from functools import partial, reduce
from json.decoder import JSONDecodeError
from smstats import custom

deref = partial(reduce, lambda dictionary, key: dictionary[key])

def get_resp_param(stage, response, expected_status, key):
    """
    Extract specified parameter from the JSON body of a http response
    """
    # Check if status code match
    if response.status_code != expected_status:
        raise custom.DataGetError(
            f'Error during stage: {stage}. Unexpected reponse status: {response.status_code}')

    # Check if body contains json
    try:
        body = response.json()
    except JSONDecodeError as decode_err:
        raise custom.DataGetError(
            f'Error during stage: {stage}. Could not read json from response') from decode_err

    # Get parameter
    try:
        return deref(key, body)
    except KeyError as key_err:
        raise custom.DataGetError(
            f'Error during stage: {stage}. Parameter {key} not found in received json response') \
        from key_err

def match_resp_param(response, expected_response):
    """
    Check if received response matches expected response
    """
    # Check if status code match
    if response.status_code != expected_response.status:
        return False

    # Check if body contains json
    try:
        body = response.json()
    except JSONDecodeError:
        return False

    # Get parameter
    try:
        value = deref(expected_response.key, body)
    except KeyError:
        return False

    # Match parameter value
    return value == expected_response.value



"""
Custom Types
"""
from collections import namedtuple, defaultdict
from dataclasses import dataclass

# Immutable types
ExpectedResponse = namedtuple('ExpectedResponse', ['status', 'key', 'value'])

# Mutable types
@dataclass
class PostDetails:
    """Derive summary of user post"""
    length: int = -1
    month: str = 'undefined'
    week: int = -1
    user: str = 'undefined'

@dataclass
class PostsSummary:
    """Keep summary of user posts"""
    total: int = 0
    longest: int = 0
    sum_length: int = 0

# pylint: disable=R0903
class IntervalSummary:
    """
    keep user wise summary of specific interval
    DEVELOPER NOTE: DataClass with defaultdict is less readable
    Hence used regular class, that resulted into pylint warning
    """
    def __init__(self):
        """Initialize empty summary"""
        self.users = defaultdict(PostsSummary)

# pylint: disable=R0903
class Summary:
    """Keep summary of user posts interval and user wise"""
    def __init__(self):
        """Initialize empty summary"""
        self.monthly = defaultdict(IntervalSummary)
        self.weekly = defaultdict(IntervalSummary)

# Exceptions
class DataGetError(Exception):
    """
    Exception raised when a HTTP response is not as expected
    """



"""
Get and Summarize Posts
"""
import requests
from smstats import config
from smstats import utils
from smstats.custom import Summary
from smstats import postutils

class PostManager:
    """
    Publishes information about posts
    """

    def __init__(self, settings=config.Config()):
        """
        Prepare to get posts
        """
        self.settings = settings
        self.get_token()

    def get_token(self):
        """
        Register a token with supermetrics service
        """
        response = requests.post(url=config.TOKEN_URL,
                                json={
                                    'client_id': self.settings.client_id,
                                    'email': self.settings.email,
                                    'name': self.settings.name})
        self.token = utils.get_resp_param('Get Token', response, 200, config.TOKEN_KEY)

    def get_posts(self, page):
        """
        Get a page of user posts
        """
        return requests.get(url=config.POSTS_URL,
                            params= {
                                'sl_token': self.token,
                                'page': page
                            })

    def get_posts_stats(self):
        """
        Get cumulative stats of user posts
        """
        summary = Summary()

        for page in range(1, self.settings.max_page+1):
            response = self.get_posts(page)

            # Token may have expired, try again
            if utils.match_resp_param(response, config.INVALID_TOKEN_RESPONSE):
                self.get_token()
                response = self.get_posts(page)

            posts = utils.get_resp_param('Get Posts', response, 200, config.POSTS_KEY)
            postutils.build_posts_summary(summary, posts)

        return postutils.get_stats(summary)
